!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! Wavefield propagation module
!! Written by Noha in 09-12, modified by Mandy in 11-12 for BEI
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

module prop_mod1DTK
  use sep
  use fourier_modOMP1D
  use smooth_mod
  implicit none 
 ! include "fftw3.f"

  logical, private       :: wfldout                                 ! Switch for output wavefield movie
  real, private :: dt, dtout, dt_prop,otout, velconst,tdelay                                   ! Propagation time step size
  real, private :: oz, dz, zsrc, okz, dkz, pi=4.*atan(1.)
  integer,private :: jt, jst                                    ! Ratio between dt and dt_out
  integer,private :: itdelay                               ! itdelay
  integer,private :: ntout, nt_prop, ntwin, nz, nnz, nkz             ! Size of data
  integer,private :: nop, npad                             ! order of Laplacian operator
  real, private,pointer :: v(:), v2(:)                         ! Velocity 
  real, private         ::d2t  ,d2t12                         ! coef
  real, private, pointer :: src_func(:,:)                 ! src wavelet, receiver data
  integer,private, pointer  :: src_loc(:,:)               ! source and receiver location
  real, private, pointer :: ztaper(:)                     ! Taper
  character(len=128), private    :: wfldfile_
  complex, dimension(:), pointer    :: in1(:), out1(:), cdatapad(:)
  contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
!! SUBROUTINE #1: INITIALIZE PROPAGATION
!! This subroutine is to prepare for propagation by calculating necessary parameters and coeffients
!! 1) Figure out a dt that is an integer ratio of dt_out for propagation and passes stability
!! 2) Calculate coefficients 
!! 3) Set velocity sampling, it will just be the same as in the input velocity model. 
!! 4) Other definitions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  
  subroutine init_propagate(vel,oz_in,dz_in,dst,dt_prop_in,ntout_in,dtout_in,otout_in,tdelay_in,vmax,npad_in,wfldout_in,wfldfile_in) 
    logical          :: wfldout_in
    real, target     :: vel(:) !Input velocity model
    real             :: dz_in !Sampling of input field

    real             :: dst, dtout_in,otout_in, dt_prop_in,dt_stable, val, tdelay_in  !Sampling of the data
    real             :: oz_in   !Origin of velocity model
    real             :: vmax    !the maximum velocity contained in the velocity model
    integer          :: iz,ntout_in,k,p, npad_in !integral ratios between dtout and dt, origins
  character(len=128)    :: wfldfile_in
    
    write(0,*) 'Running: init_propagate'

    ! (1) The stability condition to calculate the new dt from dtout:
 !   vmax=MAXVAL(vel)
    ntout=ntout_in
    dtout=dtout_in 
    otout=otout_in
    dt_prop=dt_prop_in
    nt_prop = ((ntout-1)*dtout)/dt_prop+2.5
    tdelay=tdelay_in
    dz=dz_in; oz=oz_in
    wfldout=wfldout_in
    wfldfile_=wfldfile_in

	k=ceiling(dtout/(0.8*dz/vmax))+1	!using ceiling to get the largest possible integer,
	!!since we need vmax*dt to be less than 0.707 (2D Courant number)
	dt_stable=dtout/k
        if (dt_prop .gt. dt_stable) write(0,*) 'Warning, dt_prop is greater than dt_stable', dt_stable, dt_prop
        dt=dt_prop
	jt=dtout/dt_prop
        jst=dst/dt
        write(0,*) 'init_', vmax, ntout, k, dtout, dt,jt, dz


    !(2) Coefficient calculation (8th order finite diff)
    d2t=dt*dt	;   d2t12 = dt*dt/12.
     	
 
    !(3) Find the velocity parameters
    v=>vel      

   !(4) Other defn
   nop=4;   ! Using 8th order Laplacian

   nz=size(v,1);
   nnz=nz-2*npad

   !(5) Setup damping boundary
   npad=npad_in
 !  call init_taperCos(npad)
   
  
   ! (7) square the velocity
   allocate(v2(nz))

      do iz=1,nz
         v2(iz) = v(iz)*v(iz)
!	 if (mod(iz,50) .eq. 0) write(0,*) 'iz, v2', iz, v2(iz)
      end do



  !! Initialize Fourier Transform parameter
  allocate(in1(5*nz),out1(5*nz),cdatapad(5*nz))
  nkz = 5*nz
  okz = -1.*pi/(dz)
  dkz = 2.*pi/(5*nz*dz)
  call ft_init1D(5*nz,in1,out1)

  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #2: INITIALIZE THE SOURCE
!! This subroutine define the parameters necessary to propagate source wavefields 
!! 1) Allocate src_loc(1,1): the size of this array is 1 row by 1 column
!!    Allocate src_func(ntout*jst,1) - store the source-wavelet in time
!! 2) Set src_loc location
!! 3) Interpolate source to src_func to fit the number of steps we want to propagate
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine source_init(source,zsrc_in,tdelay)
    real, allocatable, dimension(:) :: source !Source function 
    real :: zsrc_in !Source location in initial velocity model coordinate system !!
    integer :: i1,i0 !number of sources
    real :: fract, tdelay


    write(0,*) 'Running: source_init'

   !(0) Itdelay
    itdelay = (tdelay)/dt + 0.5
    zsrc=zsrc_in
    !(1) Allocations:
    allocate(src_loc(1,1), src_func(ntout*jt+itdelay,1))

    !(2) Set src_loc
    src_loc(1,1)=(zsrc-oz)/dz+.5 
    ! error check to make sure inside grid
    if ((src_loc(1,1) .gt. nz) .and. (src_loc(1,1) .lt. 1)) call erexit('src z location out of bound')

    write(0,*) 'src_loc', src_loc(1,1), zsrc, dz
 !   call erexit('exit noww')

    !(3) Interpolate source to src_func to fit the number of steps in propagation   
    src_func=0
    ! linear interpolation
    do i0=1,size(source)
       do i1=1,jst
          fract=real(i1-1)/real(jst)
          src_func(i1+(i0-1)*jst,1)=(1.-fract)*source(i0)+fract*source(min(i0+1,size(source)))
       end do
    end do



  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!! SUBROUTINE #3: RECEIVER INITIALIZATION 
!! Initialize the parameter necessary for receivers wavefield or location
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine receiver_init(ntwin_in,velconst_in)
   integer :: ntwin_in
   real    :: velconst_in 
   write(0,*) 'Running: receiver_init'
   ntwin=ntwin_in
   velconst=velconst_in

 end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #4: WAVE-FIELD PROPAGATION
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t)
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield(prev,cur,next,wave_func,wave_loc,time_marker,nt)
    real, pointer, dimension(:)    :: prev,cur,next !Previous,current, and next wavefields
    real, pointer                    :: wave_func(:,:)
    integer, pointer                 :: wave_loc(:,:)
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir,irx
    real :: val

!   write(0,*) 'Running: propagate_wavefield TK', time_marker ,'out of', jt*ntout, sum(prev), sum(cur), sum(next)
!   if (time_marker .gt. 1000) call erexit('exit nnoww')
  if (mod(time_marker,200) == 0) write(0,*) 'Running: propagate_wavefield TK', time_marker ,'out of', ntout*jt , sum(prev), sum(cur), sum(next)

    do it=1,nt
       val = 0.
        !(1) source/receiver injection
        do ir=1,size(wave_loc,2)
 !          if ((mod(it+time_marker,100).eq.1) .and. (it .lt. 2000)) write(0,*) 'wave_func', it, time_marker, wave_func(it+time_marker,ir)
           cur(wave_loc(1,ir))=cur(wave_loc(1,ir))+wave_func(it+time_marker,ir)
        end do       
       
       ! (2) Apply Laplacian;  next = Laplacian cur
!        write(0,*) 'cur before', sum(cur)
        call ApplyLaplac(cur,next)
!        write(0,*) 'cur after', sum(cur)
 


       ! (3) Time-marching 2nd order
         do iz=1,nz
            val = next(iz)
            next(iz) = d2t*v2(iz)*val + -prev(iz) + 2*cur(iz)
         end do

!      ! (3) Applying damping boundary
 !       call taperingCos(prev,cur,next)

        !(3) Time stepping
        prev=cur
        cur=next
        
        
   end do   
end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #5: WAVE-FIELD PROPAGATION 4th order in time
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t) - (dt^2)(v^2)/12 (Laplacian)(v^2 Laplacian u(t))
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield4th(prev,cur,next,next2,wave_func,wave_loc,time_marker,nt)
    real, pointer, dimension(:)    :: prev,cur,next,next2 !Previous,current, and next wavefields
    real, pointer                    :: wave_func(:,:)
    integer, pointer                 :: wave_loc(:,:)
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir,irx
    real :: val

!   write(0,*) 'Running: propagate_wavefield TK', time_marker ,'out of', jt*ntout, sum(prev), sum(cur), sum(next)
!   if (time_marker .gt. 1000) call erexit('exit nnoww')
  if (mod(time_marker,200) == 0) write(0,*) 'Running: propagate_wavefield TK', time_marker ,'out of', ntout*jt , sum(prev), sum(cur), sum(next)

    do it=1,nt
       val = 0.
        !(1) source/receiver injection
        do ir=1,size(wave_loc,2)
 !          if ((mod(it+time_marker,100).eq.1) .and. (it .lt. 2000)) write(0,*) 'wave_func', it, time_marker, wave_func(it+time_marker,ir)
           cur(wave_loc(1,ir))=cur(wave_loc(1,ir))+wave_func(it+time_marker,ir)
        end do       
       
       ! (2) Apply Laplacian;  next = Laplacian cur
!        write(0,*) 'cur before', sum(cur)
        call ApplyLaplac(cur,next)

        ! (2b) Calculate (Laplacian) (dt2 v2 Laplacian u(t) )
        do iz=1,nz
             val = next(iz)
             next(iz) = d2t*v2(iz)*val
             next2(iz) = 0.
        end do 
        call ApplyLaplac(next,next2)

       ! (3) Time-marching 2nd order
         do iz=1,nz
            val = next2(iz)
            next(iz) = -prev(iz) + 2*cur(iz) + next(iz) + d2t12*v2(iz)*val
         end do

!      ! (3) Applying damping boundary
 !       call taperingCos(prev,cur,next)

        !(3) Time stepping
        prev=cur
        cur=next
        
        
   end do   
end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #6: Propagate and save source wavefield 4th order in time
!! (1) Allocation
!! (2) propagate source wavefield
!! (3) Save to array at every jt time-step
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

  subroutine create_source_array4th(array)
  real, allocatable, dimension(:,:)  :: array !Array of nz,ntout
  integer :: iz, i, it_cur
  real, pointer, dimension(:) :: prev,cur,next,next2

  write(0,*) 'Running: create_source_array',  itdelay, nt_prop, size(array,2)
  ! (1) Allocation  
  allocate(prev(nz),cur(nz),next(nz),next2(nz))

  do iz=1,nz
     prev(iz) = 0.; 
     cur(iz) = 0.;
     next(iz) = 0.;
     next2(iz) = 0.;
  end do
 ! write(0,*) 'ntout,jt,dt', ntout, jt, dt

  do i=1,nt_prop+itdelay
  !do i=1,ntout+itdelay/jt
  !   it_cur = (i-1)*jt+1
      it_cur = i
     ! (2) Propagate source wavefield for jt time-steps
     call propagate_wavefield4th(prev,cur,next,next2,src_func,src_loc,it_cur,1)

     if (i .lt. size(array,2)) array(:,i)=next

     ! (3) Save a time-slice of sou wavefield to the array at every jt time-step
!    if (i .gt. itdelay/jt) then
!        array(:,i-itdelay/jt)=next
!     end if
  end do
    
  deallocate(prev,cur,next)

end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #8: Propagate and save source wavefield
!! (1) Allocation
!! (2) propagate source wavefield
!! (3) Save to array at every jt time-step
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

  subroutine create_source_array(array)
  real, allocatable, dimension(:,:)  :: array !Array of nz,ntout
  integer :: iz, i, it_cur
  real, pointer, dimension(:) :: prev,cur,next

  write(0,*) 'Running: create_source_array',  itdelay, nt_prop, size(array,2)
  ! (1) Allocation  
  allocate(prev(nz),cur(nz),next(nz))

  do iz=1,nz
     prev(iz) = 0.; 
     cur(iz) = 0.;
     next(iz) = 0.;
  end do
 ! write(0,*) 'ntout,jt,dt', ntout, jt, dt

  do i=1,nt_prop+itdelay
  !do i=1,ntout+itdelay/jt
  !   it_cur = (i-1)*jt+1
      it_cur = i
     ! (2) Propagate source wavefield for jt time-steps
     call propagate_wavefield(prev,cur,next,src_func,src_loc,it_cur,1)

     if (i .lt. size(array,2)) array(:,i)=next

     ! (3) Save a time-slice of sou wavefield to the array at every jt time-step
!    if (i .gt. itdelay/jt) then
!        array(:,i-itdelay/jt)=next
!     end if
  end do
    
  deallocate(prev,cur,next)

end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #12: MODEL A SHOT
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine model_shot(dat,order) 

  integer              :: itout, iz, itb, ite, it, icur,order
  real                 :: rz,tout
  real, dimension(:,:) :: dat !Shot gather
  real, allocatable, dimension(:,:) :: source_array

  write(0,*) 'Running: model_shot'
  
  ! (1) Allocation
  write(0,*) 'allocate', nz, nt_prop
  allocate(source_array(nz,nt_prop))

  ! (2) Propagate and save source wavefield
  if (order .eq. 2) then  ! 2nd order in time
      call create_source_array(source_array)
  else if (order .eq. 4) then  ! 4th order in time
      call create_source_array4th(source_array)
  end if

  if (wfldout) then
    call sep_write(source_array,wfldfile_)
    call auxputch("n1",'i',nz,wfldfile_)
    call auxputch("n2",'i',nt_prop,wfldfile_)
    call auxputch("d1",'f',dz,wfldfile_)
    call auxputch("d2",'f',dt,wfldfile_)
    call auxputch("o1",'f',oz,wfldfile_)
    call auxputch("o2",'f',otout-tdelay,wfldfile_)
    call auxputch("label1",'s','z(m)',wfldfile_) 
    call auxputch("label2",'s','t(s)',wfldfile_) 
  end if 

 !  write(0,*) 'zsrc/vel', zsrc/velconst, tdelay      

  ! (3) collect data
  do itout=1,ntout
     tout= (itout-1)*dtout+otout+1.
     rz = tout*velconst+zsrc
     iz = (rz-oz)/dz+1.5
     itb = (tout-2*tdelay-otout)/dt+.5
     ite = itb + (ntwin-1)
    ! write(0,*) '@@@',itout,iz,itb , tout, rz
    ! call erexit('exit not')
     if ( ite .lt. itb) call erexit('exit now ite less than itb')
     if ((iz .gt. nz) .or. (iz .lt. 1)) then
         write(0,*)'itout tout', itout, tout, itb, ite, iz
         dat(:,itout) = 0.
     else
        do it=1,ntwin    
           icur = itb+(it-1)
           if ((icur .lt. 1) .or. (icur .gt. nt_prop)) then
              dat(it,itout) = 0.
           else
              dat(it,itout) = source_array(iz,icur)
           end if
        end do
     end if 
  end do

  write(0,*) 'sum dat, source_array', sum(dat), sum(source_array)

  deallocate(source_array) 
  call clean_up_array()

end subroutine  



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #14: CLEANUP! 
!! Deallocate memory specific to a shot
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine clean_up_array()
     deallocate (src_loc,src_func) 	
 end subroutine
 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # 16: Apply tapering
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine taperingCos(prev,cur,next)

    integer  :: iz
    real, pointer, dimension(:) :: prev,cur,next !Previous,current, and next wavefields

          prev(1:npad) = prev(1:npad) * ztaper(1:npad:+1);
          prev(nz-npad+1:nz) = prev(nz-npad+1:nz) * ztaper(npad:1:-1)
          cur(1:npad) = cur(1:npad) * ztaper(1:npad:+1);
          cur(nz-npad+1:nz) = cur(nz-npad+1:nz) * ztaper(npad:1:-1)
          next(1:npad) = next(1:npad) * ztaper(1:npad:+1);
          next(nz-npad+1:nz) = next(nz-npad+1:nz) * ztaper(npad:1:-1)




  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 17: Tapering Setup
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
    subroutine init_taperCos(npad)

    integer                     :: iz, ix,ii, npad
    real                       :: dist
    real                        :: pi=3.14159265359


    allocate(ztaper(npad))

       ! all four edges
       do ii=1,npad
          ztaper(ii) = 0.985 + 0.015*cos(pi*(npad-ii)/npad)
       end do


  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 21: Take derivative for 1D array
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)
  end subroutine 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 23: Take Vertical Derivative of image
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
  subroutine Ver2ndDer(adj,m,nz)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nz,iz
  real, dimension(:), allocatable :: temp
  real, dimension(:)      :: m


  allocate(temp(nz))

      do iz=2,nz-1
         if (adj) then
            m(iz) = m(iz) + (-2.*temp(iz) + temp(iz+1) + temp(iz-1))
         else
            m(iz) = m(iz) + -2*temp(iz)
            m(iz) = m(iz) + temp(iz+1)
            m(iz) = m(iz) + temp(iz-1)
         end if
       end do


  end subroutine



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # XX: Apply Laplacian 
!! 1) Fourier transform
!! 2) multiply by -kz**2
!! 3) Inverse fourier transform
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine ApplyLaplac(inArr,outArr)

  integer               :: ikz,iz
  real                  :: kz
  real, dimension(:)    :: inArr, outArr
  complex               :: ckz
 
  !write(0,*) 'Inside Apply Laplac'
  !write(0,*) 'nz, dz', nz, dz
  !write(0,*) 'nkz, okz, dkz', nkz, okz, dkz
  

  !! Forward transform
   do iz=1,nz
      cdatapad(iz) = cmplx(inArr(iz),0.)
   end do
   call ft1axis(.false.,-1.,cdatapad,in1,out1)

  !! Apply Laplacian
  do ikz=1,nkz
     kz = okz + (ikz-1)*dkz
     ckz = cmplx(kz,0.)
 !    if (mod(ikz,100) .eq. 1) write(0,*) 'ikz, kz, ckz', ikz, kz, ckz
     cdatapad(ikz) = cdatapad(ikz)*kz*kz*-1.
  end do


  !! Inverse transform
  call ft1axis(.true.,-1.,cdatapad,in1,out1)
  do iz=1,nz
     outArr(iz) = real(cdatapad(iz))
  end do

!  write(0,*) 'inArr', inArr(1:100)
!  write(0,*) 'outArr', outArr(1:100)
!  call erexit('exit now')

end subroutine


end module


