!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! Wavefield propagation module
!! Written by Noha in 09-12, modified by Mandy in 11-12 for BEI
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

module prop_mod2DTK
  use sep
  use fourier_modOMP
  use smooth_mod
  implicit none 
 ! include "fftw3.f"

  logical, private       :: wfldout                                 ! Switch for output wavefield movie
  real, private :: dt, dtrec, dt_prop,otrec, velconst,tdelay                                   ! Propagation time step size
  real, private :: oz, dz, ox, dx, zsrc, orx, drx, okz, dkz, okx, dkx, pi=4.*atan(1.)
  integer,private :: jt, jst                                    ! Ratio between dt and dt_out
  integer,private :: itdelay                               ! itdelay
  integer,private :: ntrec, nt_prop, ntwin
  integer,private :: nx, nnx, nkx, nz, nnz, nkz               ! Size of data
  integer,private :: nop, npad, nrx                                ! Order of Laplacian operator
  real, private,pointer :: v(:,:), v2(:,:)                    ! Velocity 
  real, private         ::d2t  ,d2t12                         ! coef
  real, private, pointer :: src_func(:,:)                     ! src wavelet, receiver data
  integer,private, pointer  :: src_loc(:,:), rec_loc(:,:)     ! source and receiver location
  real, private, pointer :: zxtaper(:),ctaper(:,:)       ! Taper
  character(len=128), private    :: wfldfile_
  complex, dimension(:), pointer    :: in1, out1, in2, out2
  complex, dimension(:,:), pointer  ::  cdatapad
  contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
!! SUBROUTINE #1: INITIALIZE PROPAGATION
!! This subroutine is to prepare for propagation by calculating necessary parameters and coeffients
!! 1) Figure out a dt that is an integer ratio of dt_out for propagation and passes stability
!! 2) Calculate coefficients 
!! 3) Set velocity sampling, it will just be the same as in the input velocity model. 
!! 4) Other definitions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
  
  subroutine init_propagate(vel,oz_in,dz_in,ox_in,dx_in,dst,dt_prop_in,ntrec_in,dtrec_in,otrec_in,tdelay_in,vmax,npad_in,wfldout_in,wfldfile_in) 
    logical          :: wfldout_in
    real, target     :: vel(:,:) !Input velocity model
    real             :: dz_in, dx_in !Sampling of input field

    real             :: dst, dtrec_in,otrec_in, dt_prop_in,dt_stable, val, tdelay_in  !Sampling of the data
    real             :: oz_in, ox_in   !Origin of velocity model
    real             :: vmax    !the maximum velocity contained in the velocity model
    integer          :: iz, ix,ntrec_in,k,p, npad_in !integral ratios between dtrec and dt, origins
  character(len=128)    :: wfldfile_in
    
    write(0,*) 'Running: init_propagate'

    ! (1) The stability condition to calculate the new dt from dtrec:
 !   vmax=MAXVAL(vel)
    ntrec=ntrec_in
    dtrec=dtrec_in 
    otrec=otrec_in
    dt_prop=dt_prop_in
    nt_prop = ((ntrec-1)*dtrec)/dt_prop+2.5
    tdelay=tdelay_in
    dz=dz_in; oz=oz_in
    dx=dx_in; ox=ox_in;
    wfldout=wfldout_in
    wfldfile_=wfldfile_in

	k=ceiling(dtrec/(0.8*dz/vmax))+1	!using ceiling to get the largest possible integer,
	!!since we need vmax*dt to be less than 0.707 (2D Courant number)
	dt_stable=dtrec/k
        if (dt_prop .gt. dt_stable) write(0,*) 'Warning, dt_prop is greater than dt_stable', dt_stable, dt_prop
        dt=dt_prop
	jt=dtrec/dt_prop
        jst=dst/dt
        write(0,*) 'init_', vmax, ntrec, k, dtrec, dt,jt, dz


    !(2) Coefficient calculation (8th order finite diff)
    d2t=dt*dt	;   d2t12 = dt*dt/12.
     	
 
    !(3) Find the velocity parameters
    v=>vel      

   !(4) Other defn
   nop=4;   ! Using 8th order Laplacian

   nz=size(v,1);
   nnz=nz-2*npad
   nx=size(v,2);
   nnx=nx

   !(5) Setup damping boundary
   npad=npad_in
   call init_taperCos(npad)
   
  
   ! (7) square the velocity
   allocate(v2(nz,nx))

   do ix=1,nx
      do iz=1,nz
         v2(iz,ix) = v(iz,ix)*v(iz,ix)
!	 if (mod(iz,50) .eq. 0) write(0,*) 'iz, v2', iz, v2(iz)
      end do
   end do

  !! Initialize Fourier Transform parameter
  allocate(in1(2*nz),out1(2*nz),cdatapad(2*nz,2*nx))
  allocate(in2(2*nx),out2(2*nx))
  nkz = 2*nz
  okz = -1.*pi/(dz)
  dkz = 2.*pi/(2*nz*dz)
  nkx = 2*nx
  okx = -1.*pi/(dx)
  dkx = 2.*pi/(2*nx*dx)
  call ft_init(2*nz, 2*nx,in1,out1,in2,out2)

  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #2: INITIALIZE THE SOURCE
!! This subroutine define the parameters necessary to propagate source wavefields 
!! 1) Allocate src_loc(1,1): the size of this array is 1 row by 1 column
!!    Allocate src_func(ntrec*jst,1) - store the source-wavelet in time
!! 2) Set src_loc location
!! 3) Interpolate source to src_func to fit the number of steps we want to propagate
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  subroutine source_init(source,zsrc_in,xsrc_in,tdelay)
    real, allocatable, dimension(:) :: source !Source function 
    real :: zsrc_in, xsrc_in !Source location in initial velocity model coordinate system !!
    integer :: i1,i0 !number of sources
    real :: fract, tdelay


    write(0,*) 'Running: source_init'

   !(0) Itdelay
    itdelay = (tdelay)/dt + 0.5
    zsrc=zsrc_in
    !(1) Allocations:
    allocate(src_loc(2,1), src_func(ntrec*jt+itdelay,1))

    !(2) Set src_loc
    src_loc(1,1)=(zsrc_in-oz)/dz+.5 +max(nop,npad)
    src_loc(2,1)=(xsrc_in-ox)/dx+.5 +nop

    ! error check to make sure inside grid
    if ((src_loc(1,1) .gt. nz) .and. (src_loc(1,1) .lt. 1)) call erexit('src z location out of bound')
    if ((src_loc(2,1) .gt. nnx+npad) .and. (src_loc(2,1) .lt. 1)) call erexit('src x location out of bound')


    !(3) Interpolate source to src_func to fit the number of steps in propagation   
    src_func=0
    ! linear interpolation
    do i0=1,size(source)
       do i1=1,jst
          fract=real(i1-1)/real(jst)
          src_func(i1+(i0-1)*jst,1)=(1.-fract)*source(i0)+fract*source(min(i0+1,size(source)))
       end do
    end do



  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!! SUBROUTINE #3: RECEIVER INITIALIZATION 
!! Initialize the parameter necessary for receivers wavefield or location
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine receiver_init(zrec,nrx_in,orx_in,drx_in)
   real    :: zrec  !Position of the source
   real    :: orx_in,drx_in !Sampling of the receivers compared to the source position
   integer :: nrx_in
   integer :: ir,irx
   real    :: xr
   

   nrx = nrx_in
   orx = orx_in
   drx = drx_in

   write(0,*) 'Running: receiver_init'
   ! write(0,*) 'xsrc, zsrc', xsrc, zsrc
   
   !(1) Allocation
   allocate (rec_loc(2,nrx))   

   !(2) Set rec_loc 
   do irx=1,nrx
     xr = (irx-1)*drx+orx
     rec_loc(2,irx)=(xr-ox)/dx+0.5+nop
   end do
   rec_loc(1,:)=(zrec-oz)/dz+0.5 + max(nop,npad)
   write(0,*) 'irz', rec_loc(1,1), zrec
   !check to make sure inside grid
   if ( (rec_loc(1,1) .lt. 1) .or. (rec_loc(1,1) .gt. nz)) call erexit('receiver z location out of bound')
   write(0,*) 'rec_loc', rec_loc(2,:)

   do irx=1,nrx
      ir = rec_loc(2,irx)
      if ( (ir .lt. 1) .or. (ir .gt. nnx+npad)) call erexit('receiver x location out of bound')
   end do

 end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #4: WAVE-FIELD PROPAGATION
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t)
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield(prev,cur,next,wave_func,wave_loc,time_marker,nt)
    real, pointer, dimension(:,:)    :: prev,cur,next !Previous,current, and next wavefields
    real, pointer                    :: wave_func(:,:)
    integer, pointer                 :: wave_loc(:,:)
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir
    real :: val

!   write(0,*) 'Running: propagate_wavefield TK', time_marker ,'out of', jt*ntrec, sum(prev), sum(cur), sum(next)
!   if (time_marker .gt. 1000) call erexit('exit nnoww')
  if (mod(time_marker,200) == 0) write(0,*) 'Running: propagate_wavefield TK', time_marker ,'out of', ntrec*jt , sum(prev), sum(cur), sum(next)

    do it=1,nt
       val = 0.
        !(1) source/receiver injection
        do ir=1,size(wave_loc,2)
 !          if ((mod(it+time_marker,100).eq.1) .and. (it .lt. 2000)) write(0,*) 'wave_func', it, time_marker, wave_func(it+time_marker,ir)
           cur(wave_loc(1,ir),wave_loc(2,ir))=cur(wave_loc(1,ir),wave_loc(2,ir))+wave_func(it+time_marker,ir)
        end do       
       
       ! (2) Apply Laplacian;  next = Laplacian cur
!        write(0,*) 'cur before', sum(cur)
        call ApplyLaplac2D(cur,next)
!        write(0,*) 'cur after', sum(cur)
 


       ! (3) Time-marching 2nd order
         do ix=1,nx
           do iz=1,nz
            val = next(iz,ix)
            next(iz,ix) = d2t*v2(iz,ix)*val + -prev(iz,ix) + 2*cur(iz,ix)
           end do
         end do

!      ! (3) Applying damping boundary
        call taperingCos(prev,cur,next)

        !(3) Time stepping
        prev=cur
        cur=next
        
        
   end do   
end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #5: WAVE-FIELD PROPAGATION 4th order in time
!! Input: prev,cur,next =  u(t+1), u(t), u(t-1)
!! Input: rec_f = time-index marker on where to inject data
!! Input: nt = number of time step to propagate
!! output: next (represent the wavefield after propagating nt time steps)
!! (1) Apply Laplacian and time-marching (in reverse time)
!!         u(t+/-1) = - u(t-/+1) + 2*u(t) + (v^2)*(dt^2)*(Laplacian)u(t) - (dt^2)(v^2)/12 (Laplacian)(v^2 Laplacian u(t))
!! Note that propagate forward and reverse in time is symmetric
!! (2) data/source injection       
!! (3) Time stepping (update the 3 time slices)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 subroutine propagate_wavefield4th(prev,cur,next,next2,wave_func,wave_loc,time_marker,nt)
    real, pointer, dimension(:,:)    :: prev,cur,next,next2 !Previous,current, and next wavefields
    real, pointer                    :: wave_func(:,:)
    integer, pointer                 :: wave_loc(:,:)
    integer :: time_marker             !Current time step marker for injection
    integer :: ix,iz,it,nt,ir
    real :: val

!   write(0,*) 'Running: propagate_wavefield TK', time_marker ,'out of', jt*ntrec, sum(prev), sum(cur), sum(next)
!   if (time_marker .gt. 1000) call erexit('exit nnoww')
  if (mod(time_marker,200) == 0) write(0,*) 'Running: propagate_wavefield TK', time_marker ,'out of', ntrec*jt , sum(prev), sum(cur), sum(next)

    do it=1,nt
       val = 0.
        !(1) source/receiver injection
        do ir=1,size(wave_loc,2)
 !          if ((mod(it+time_marker,100).eq.1) .and. (it .lt. 2000)) write(0,*) 'wave_func', it, time_marker, wave_func(it+time_marker,ir)
           cur(wave_loc(1,ir),wave_loc(2,ir))=cur(wave_loc(1,ir),wave_loc(2,ir))+wave_func(it+time_marker,ir)
        end do       
       
       ! (2) Apply Laplacian;  next = Laplacian cur
!        write(0,*) 'cur before', sum(cur)
        call ApplyLaplac2D(cur,next)

        ! (2b) Calculate (Laplacian) (dt2 v2 Laplacian u(t) )
        do ix=1,nx
          do iz=1,nz
             val = next(iz,ix)
             next(iz,ix) = d2t*v2(iz,ix)*val
             next2(iz,ix) = 0.
          end do
        end do 
        call ApplyLaplac2D(next,next2)

       ! (3) Time-marching 2nd order
         do ix=1,nx
           do iz=1,nz
              val = next2(iz,ix)
              next(iz,ix) = -prev(iz,ix) + 2*cur(iz,ix) + next(iz,ix) + d2t12*v2(iz,ix)*val
           end do
         end do

!      ! (3) Applying damping boundary
        call taperingCos(prev,cur,next)

        !(3) Time stepping
        prev=cur
        cur=next
        
        
   end do   
end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #6: Propagate and save source wavefield 4th order in time
!! (1) Allocation
!! (2) propagate source wavefield
!! (3) Save to array at every jt time-step
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

  subroutine create_source_array4th(array)
  real, allocatable, dimension(:,:,:)  :: array !Array of nz,nx,ntrec
  integer :: iz,ix, i, it_cur, itrec
  real    :: t
  real, pointer, dimension(:,:) :: prev,cur,next,next2

  write(0,*) 'Running: create_source_array',  itdelay, nt_prop, size(array,2)
  ! (1) Allocation  
  allocate(prev(nz,nx),cur(nz,nx),next(nz,nx),next2(nz,nx))

  do ix=1,nx
    do iz=1,nz
       prev(iz,ix) = 0.; 
       cur(iz,ix) = 0.;
       next(iz,ix) = 0.;
       next2(iz,ix) = 0.;
    end do
 end do
 ! write(0,*) 'ntrec,jt,dt', ntrec, jt, dt

  do i=1,nt_prop+itdelay
  !do i=1,ntrec+itdelay/jt
  !   it_cur = (i-1)*jt+1
      it_cur = i
     ! (2) Propagate source wavefield for jt time-steps
     call propagate_wavefield4th(prev,cur,next,next2,src_func,src_loc,it_cur,1)

     t=(i-1)*dt_prop+otrec
     if (jt .gt. 1) then
       if ( (t-tdelay .ge. 0.) .and. (mod(i-itdelay,jt) .eq. 1) ) then
          itrec = (i-itdelay)/jt + 1
   !      write(0,*) 'i,itrec,t',i,itrec,t
          array(:,:,itrec)=next
       end if
     else
       if ( (t-tdelay .ge. 0.)) then
          itrec = (i-itdelay)/jt +1
!         write(0,*) 'i,itrec,t',i,itrec,size(array,3)
          if ((itrec .gt. 1) .and. (itrec .le. ntrec))  array(:,:,itrec)=next
       end if
     end if

  end do
    
  deallocate(prev,cur,next)

end subroutine


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #8: Propagate and save source wavefield
!! (1) Allocation
!! (2) propagate source wavefield
!! (3) Save to array at every jt time-step
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

  subroutine create_source_array(array)
  real, allocatable, dimension(:,:,:)  :: array !Array of nz,ntrec
  integer :: iz,ix, i, it_cur, itrec
  real    :: t
  real, pointer, dimension(:,:) :: prev,cur,next

  write(0,*) 'Running: create_source_array',  itdelay, nt_prop, size(array,2)
  ! (1) Allocation  
  allocate(prev(nz,nx),cur(nz,nx),next(nz,nx))

  do ix=1,nx
    do iz=1,nz
       prev(iz,ix) = 0.; 
       cur(iz,ix) = 0.;
       next(iz,ix) = 0.;
    end do
  end do
  write(0,*) 'ntrec,jt,dt', ntrec, jt, dt


  do i=1,nt_prop+itdelay
  !do i=1,ntrec+itdelay/jt
  !   it_cur = (i-1)*jt+1
      it_cur = i
     ! (2) Propagate source wavefield for jt time-steps
     call propagate_wavefield(prev,cur,next,src_func,src_loc,it_cur,1)
 
    t=(i-1)*dt_prop+otrec
    if (jt .gt. 1) then
       if ( (t-tdelay .ge. 0.) .and. (mod(i-itdelay,jt) .eq. 1) ) then
          itrec = (i-itdelay)/jt+1
!          write(0,*) 'i,itrec,t',i,itrec,t
          array(:,:,itrec)=next
       end if
    else
       if ( (t-tdelay .ge. 0.) ) then
          itrec = (i-itdelay)/jt+1
!         write(0,*) 'i,itrec,t',i,itrec,size(array,3)
          if ((itrec .gt. 1) .and. (itrec .le. ntrec))  array(:,:,itrec)=next
       end if
    end if

  end do
    
  deallocate(prev,cur,next)

end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE #12: MODEL A SHOT
!! (1) Allocation
!! (2) Propagate and save source wavefield
!! (3) Collect data
!! (4) Output source movie (if optioned)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine model_shot(dat,order,outname_) 

  integer              :: itrec, iz,ix, it, icur,order, irx
  real                 :: rz,trec
  real, dimension(:,:) :: dat !Shot gather
  real, allocatable, dimension(:,:,:) :: source_array
  character(len=128)   :: outname_

  write(0,*) 'Running: model_shot'
  
  ! (1) Allocation
  write(0,*) 'allocate', nz, nt_prop
  allocate(source_array(nz,nx,ntrec))

  ! (2) Propagate and save source wavefield
  if (order .eq. 2) then  ! 2nd order in time
      call create_source_array(source_array)
  else if (order .eq. 4) then  ! 4th order in time
      call create_source_array4th(source_array)
  end if

  if (wfldout) then
    call sep_write(source_array,wfldfile_)
    call auxputch("n1",'i',nz,wfldfile_)
    call auxputch("n2",'i',nx,wfldfile_)
    call auxputch("n3",'i',ntrec,wfldfile_)
    call auxputch("d1",'f',dz,wfldfile_)
    call auxputch("d2",'f',dx,wfldfile_)
    call auxputch("d3",'f',dtrec,wfldfile_)
    call auxputch("o1",'f',oz,wfldfile_)
    call auxputch("o2",'f',ox,wfldfile_)
    call auxputch("o3",'f',otrec,wfldfile_)
    call auxputch("label1",'s','z(m)',wfldfile_) 
    call auxputch("label2",'s','x(m)',wfldfile_) 
    call auxputch("label3",'s','t(s)',wfldfile_) 
  end if 

 !  write(0,*) 'zsrc/vel', zsrc/velconst, tdelay      

  ! (3) collect data
  do irx=1,size(dat,2)
     dat(:,irx) = source_array(rec_loc(1,irx),rec_loc(2,irx),:)  
  end do


  if (order .eq. 4) then
  ! Output data
    call auxputch("n1",'i',ntrec,outname_)
    call auxputch("n2",'i',nrx,outname_)
    call auxputch("d1",'f',dtrec,outname_)
    call auxputch("d2",'f',drx,outname_)
    call auxputch("o1",'f',otrec,outname_)
    call auxputch("o2",'f',orx,outname_)
    call auxputch("label1",'s','t(s)',outname_) 
    call auxputch("label2",'s','x(m)',outname_) 
    call sep_write(dat,outname_)
 !   write(0,*) 'order', order, outname_
 !   call erexit('exit now')
  end if

  write(0,*) 'sum dat, source_array', sum(dat), sum(source_array)

  deallocate(source_array) 
  call clean_up_array()

end subroutine  



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
!! SUBROUTINE #14: CLEANUP! 
!! Deallocate memory specific to a shot
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

 subroutine clean_up_array()
     deallocate (src_loc,src_func,rec_loc) 	
 end subroutine
 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # 16: Apply tapering
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine taperingCos(prev,cur,next)

    integer  :: iz, ix
    real, pointer, dimension(:,:) :: prev,cur,next !Previous,current, and next wavefields


       ! Corner
       do ix=1,npad
          do iz=1,npad

!              prev(iz,ix) = prev(iz,ix)*ctaper(iz,ix);                   ! top left
!              prev(iz,nx-ix+1) = prev(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
!              prev(nz-iz+1,nx-ix+1) = prev(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
!              prev(nz-iz+1,ix) = prev(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left

!              cur(iz,ix) = cur(iz,ix)*ctaper(iz,ix);                   ! top left
!              cur(iz,nx-ix+1) = cur(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
!              cur(nz-iz+1,nx-ix+1) = cur(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
!              cur(nz-iz+1,ix) = cur(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left

!              next(iz,ix) = next(iz,ix)*ctaper(iz,ix);                   ! top left
!              next(iz,nx-ix+1) = next(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
!              next(nz-iz+1,nx-ix+1) = next(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
!              next(nz-iz+1,ix) = next(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
          end do
       end do

       ! Left Right
!       do iz=npad+1,nz-npad
!          prev(iz, 1:npad) = prev(iz, 1:npad) * zxtaper(1:npad:+1); 
!          prev(iz,nx-npad+1:nx) = prev(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
!          cur(iz, 1:npad) = cur(iz, 1:npad) * zxtaper(1:npad:+1);
!          cur(iz,nx-npad+1:nx) = cur(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
!          next(iz, 1:npad) = next(iz, 1:npad) * zxtaper(1:npad:+1);
!          next(iz,nx-npad+1:nx) = next(iz,nx-npad+1:nx) * zxtaper(npad:1:-1)
!       end do

       ! Top and Bottom
       do ix=npad+1,nx-npad
          prev(1:npad,ix) = prev(1:npad,ix) * zxtaper(1:npad:+1);
          prev(nz-npad+1:nz,ix) = prev(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
          cur(1:npad,ix) = cur(1:npad,ix) * zxtaper(1:npad:+1);
          cur(nz-npad+1:nz,ix) = cur(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
          next(1:npad,ix) = next(1:npad,ix) * zxtaper(1:npad:+1);
          next(nz-npad+1:nz,ix) = next(nz-npad+1:nz,ix) * zxtaper(npad:1:-1)
       end do



  end subroutine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 17: Tapering Setup
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
    subroutine init_taperCos(npad)

    integer                     :: iz, ix,ii, npad
    real                       :: dist
    real                        :: pi=3.14159265359


    allocate(zxtaper(npad), ctaper(npad,npad))

       ! all four edges
       do ii=1,npad
          zxtaper(ii) = 0.985 + 0.015*cos(pi*(npad-ii)/npad)
       end do

       ! Corner
       do ix=1,npad
          do iz=1,npad
              dist = sqrt(1.001*(npad-ix)*(npad-ix)/npad/npad + 0.999999*(npad-iz)*(npad-iz)/npad/npad)
              ctaper(iz,ix) =  0.985 + 0.015*cos(pi*dist)
          end do
       end do

  end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 21: Take derivative for 1D array
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)
  end subroutine 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
!! SUBROUTINE # 23: Take Vertical Derivative of image
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!$
  subroutine Ver2ndDer(adj,m,nz)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nz,iz
  real, dimension(:), allocatable :: temp
  real, dimension(:)      :: m


  allocate(temp(nz))

      do iz=2,nz-1
         if (adj) then
            m(iz) = m(iz) + (-2.*temp(iz) + temp(iz+1) + temp(iz-1))
         else
            m(iz) = m(iz) + -2*temp(iz)
            m(iz) = m(iz) + temp(iz+1)
            m(iz) = m(iz) + temp(iz-1)
         end if
       end do


  end subroutine



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! SUBROUTINE # XX: Apply Laplacian 
!! 1) Fourier transform
!! 2) multiply by -kz**2
!! 3) Inverse fourier transform
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine ApplyLaplac2D(inArr,outArr)

  integer               :: ikz,iz, ikx, ix
  real                  :: kz, kx
  real, dimension(:,:)    :: inArr, outArr
  complex               :: ckz, ckx
 
  !write(0,*) 'Inside Apply Laplac'
  !write(0,*) 'nz, dz', nz, dz
  !write(0,*) 'nkz, okz, dkz', nkz, okz, dkz
  

  !! Forward transform
   do ix=1,nx
   do iz=1,nz
      cdatapad(iz,ix) = cmplx(inArr(iz,ix),0.)
   end do
   end do
   call ft1axis(.false.,-1.,cdatapad,in1,out1)
   call ft2axis(.false.,-1.,cdatapad,in2,out2)

  !! Apply Laplacian
  do ikx=1,nkx
     kx = okx + (ikx-1)*dkx
  do ikz=1,nkz
     kz = okz + (ikz-1)*dkz
 !    ckz = cmplx(kz,0.)
 !    if (mod(ikz,100) .eq. 1) write(0,*) 'ikz, kz, ckz', ikz, kz, ckz
     cdatapad(ikz,ikx) = cdatapad(ikz,ikx)*( kz*kz*-1. + kx*kx*-1.)
  end do
  end do


  !! Inverse transform
  call ft2axis(.true.,-1.,cdatapad,in2,out2)
  call ft1axis(.true.,-1.,cdatapad,in1,out1)

  do ix=1,nx
  do iz=1,nz
     outArr(iz,ix) = real(cdatapad(iz,ix))
  end do
  end do

!  write(0,*) 'inArr', inArr(1:100)
!  write(0,*) 'outArr', outArr(1:100)
!  call erexit('exit now')

end subroutine


end module


